/*
 * Newt-Puzzle
 * Copyright (C) Aude Jaulent 2008, 2009 <elderme@tuxfamily.org>
 * 
 * Newt-Puzzle is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Newt-Puzzle is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "display3D.h"

Display3D::Display3D(): lightKey(0), mpSky(NULL), mpFog(NULL), mpWaterPlane(NULL), mpSkyPlane(NULL), mpFBO(NULL), mpShadowMap(NULL),
						mGroundColor(Color::BLACK), mSkyColor(Color::BLACK), mHemiAxis(Vector3(0.,1.,0.)), mMaxLightSources(-1) {
	//mpFBO = new FrameBufferObject (1024, 1024);
}

Display3D::~Display3D() {
	delete mpFog;
	delete mpCamera;
	delete mpSky;
	delete mpShadowMap;
	delete mpSkyPlane;
	delete mpWaterPlane;	
	delete mpFBO;
		
	for(int i=0;i<entityTree.size();i++) {
		if(entityTree[i] != NULL) {
			delete entityTree[i];
		}
	}
	entityTree.clear();

	
	for(int i=0;i<mirrorList.size();i++) {
		delete mirrorList[i];
	}
	mirrorList.clear();
	
	for(int i=0;i<lightsD.size();i++) {
		delete lightsD[i];
	}
	lightsD.clear();
	
	for(int i=0;i<lightsS.size();i++) {
		delete lightsS[i];
	}
	lightsS.clear();
	
	for(int i=0;i<lightsP.size();i++) {
		delete lightsP[i];
	}
	lightsP.clear();
	
}

void Display3D::setRenderingMode (renderingMode rendering) {
	mRendering = rendering;	
}

void Display3D::setMaxLightSources (int maxLightSources) {
	mMaxLightSources = maxLightSources;
}

renderingMode Display3D::getRenderingMode () {
	return mRendering;
}


void Display3D::setCamera(Camera *camera) {
	mpCamera = camera;
}

void Display3D::setFog(Fog *fog) {
	mpFog = fog;
}
	
void Display3D::setSky(Sky *sky) {
	mpSky = sky;	
}

void Display3D::setSkyPlane(SkyPlane *skyPlane) {
	mpSkyPlane = skyPlane;
}

void Display3D::setWaterPlane(WaterPlane *waterPlane) {
	mpWaterPlane = waterPlane;
}


Camera* Display3D::getCamera() {
	return mpCamera;
}

Fog* Display3D::getFog() {
	return mpFog;
}

Sky* Display3D::getSky() {
	return mpSky;
}


void Display3D::setAmbientColors(Color groundColor, Color skyColor, Vector3 hemiAxis) {
	mGroundColor = groundColor;
	mSkyColor = skyColor;
	
	Vector3 xCam = mpCamera->getXAxis();
	Vector3 yCam = mpCamera->getYAxis();
	Vector3 zCam = mpCamera->getZAxis();
	
	//mHemiAxis has to take the eye position into account
	
	mHemiAxis.x = -1. * hemiAxis * (mpCamera->getXAxis());
	mHemiAxis.y = hemiAxis * (mpCamera->getYAxis());
	mHemiAxis.z = -1. * hemiAxis * (mpCamera->getZAxis());
	
	mHemiAxis.normalize ();
		
	
	if(mRendering == FIXED_PIPELINE_RENDERING) {
		Color ambientModel = Color(0.5*(groundColor.r + skyColor.r),
								   0.5*(groundColor.g + skyColor.g),
								   0.5*(groundColor.b + skyColor.b),
								   0.5*(groundColor.a + skyColor.a));
		glLightModelfv (GL_LIGHT_MODEL_AMBIENT, ambientModel);
	} 
}


void Display3D::addEntity(Entity *entity) {
	entityTree.push_back(entity);
}

void Display3D::addMirror(Mirror *mirror) {
	mirrorList.push_back(mirror);
}


void Display3D::addLightDir(LightDir *lightD) {
	int nbLights = lightsD.size() + lightsS.size() + lightsP.size();
	if((nbLights < mMaxLightSources)||(mMaxLightSources<=0)) {
		lightsD.push_back(lightD);
	}
}

void Display3D::addLightSpot(LightSpot *lightS) {
	int nbLights = lightsD.size() + lightsS.size() + lightsP.size();
	if((nbLights < mMaxLightSources)||(mMaxLightSources<=0)) {	
		lightsS.push_back(lightS);
	}
}

void Display3D::addLightPoint(LightPoint *lightP) {
	int nbLights = lightsD.size() + lightsS.size() + lightsP.size();
	if((nbLights < mMaxLightSources)||(mMaxLightSources<=0)) {	
		lightsP.push_back(lightP);
	}
}


int Display3D::getLightKey() {
	return lightKey;
}


void Display3D::activateLights() {
	glEnable(GL_LIGHTING);
	int nbD = 0;
	int nbS = 0;
	int nbP = 0;

	
	for(int i=0;i<lightsD.size();i++) {
		if(lightsD[i]->getStatus()) {
			lightsD[i]->activate(i);
			nbD++;
		}
	}
	for(int i=0;i<lightsS.size();i++) {
		if(lightsS[i]->getStatus()) {
			lightsS[i]->activate(i+nbD);
			nbS++;
		}
	}	
	for(int i=0;i<lightsP.size();i++) {
		if(lightsP[i]->getStatus()) {
			lightsP[i]->activate(i+nbD+nbS);
			nbP++;
		}
	}	
	
	lightKey = nbP + 8*nbS + 64*nbD;
	
	shaderMana->setLightKey(lightKey);
}

void Display3D::releaseLights() {

	for(int i=0;i<lightsD.size();i++) {
		lightsD[i]->release();
	}
	for(int i=0;i<lightsS.size();i++) {
		lightsS[i]->release();
	}
	for(int i=0;i<lightsP.size();i++) {
		lightsP[i]->release();
	}	
	glDisable(GL_LIGHTING);	
}


std::set<int> Display3D::getMaterialKeys() {
	return materialKeys;
}


int Display3D::getMinDirLights() {
	int res = 0;
	for(int i=0;i<lightsD.size();i++) {
		if(lightsD[i]->getStatus() == 2) {
			res++;
		}
	}
	return res;
}

int Display3D::getMaxDirLights() {
	return lightsD.size();
}

int Display3D::getMinSpotLights() {
	int res = 0;
	for(int i=0;i<lightsS.size();i++) {
		if(lightsS[i]->getStatus() == 2) {
			res++;
		}
	}
	return res;
}

int Display3D::getMaxSpotLights() {
	return lightsS.size();
}

int Display3D::getMinPointLights() {
	int res = 0;
	for(int i=0;i<lightsP.size();i++) {
		if(lightsP[i]->getStatus() == 2) {
			res++;
		}
	}
	return res;
}

int Display3D::getMaxPointLights() {
	return lightsP.size();
}


void Display3D::init(textureMode texturing) {
	
	for(int i=0;i<entityTree.size();i++) {
				
		if((mRendering!=FIXED_PIPELINE_RENDERING)&&(texturing!=TEXTURE_NONE)) {
			std::set<int> entityKeys = entityTree[i]->getMaterialKeys();

			materialKeys.insert(entityKeys.begin(), entityKeys.end());			
		}
	}
	
	for(int i=0;i<mirrorList.size();i++) {
		mirrorList[i]->initialize(NULL);

		if((mRendering!=FIXED_PIPELINE_RENDERING)&&(texturing!=TEXTURE_NONE)) {
			std::set<int> entityKeys = mirrorList[i]->getMaterialKeys();

			materialKeys.insert(entityKeys.begin(), entityKeys.end());
		}		
	}
}


void Display3D::compileShaders() {
	if(mRendering!=FIXED_PIPELINE_RENDERING) {
		if(mpSky != NULL) {
			shaderMana->compileSky();
		}
	
		if(mRendering == PER_FRAG_LIGHT_RENDERING) {
			
			std::set<int> materialKeys = getMaterialKeys();		
			
			shaderMana->compilePerFragLight(getMinDirLights(), getMaxDirLights(), 
											getMinSpotLights(), getMaxSpotLights(), 
											getMinPointLights(), getMaxPointLights(), 
											materialKeys);			
		}
	}
}


void Display3D::addMaterialsInTheSet(std::set<Material*> &matSet) {
	for(int i=0;i<entityTree.size();i++) {
		entityTree[i]->addMaterialsInTheSet(matSet);
	}
	for(int i=0;i<mirrorList.size();i++) {
		mirrorList[i]->addMaterialsInTheSet(matSet);
	}	
}

void Display3D::addMeshesInTheSet(std::set<std::pair<Mesh*, int> > &meshSet) {
	
	int nbTextures = 0;
	
	for(int i=0;i<entityTree.size();i++) {
		if(entityTree[i]->getMaterial() != NULL) {
			nbTextures = entityTree[i]->getMaterial()->getNumberOfTextures();
		}
		entityTree[i]->addMeshesInTheSet(meshSet, nbTextures);
	}
	for(int i=0;i<mirrorList.size();i++) {
		if(mirrorList[i]->getMaterial() != NULL) {
			nbTextures = mirrorList[i]->getMaterial()->getNumberOfTextures();
		}		
		mirrorList[i]->addMeshesInTheSet(meshSet, nbTextures);
	}	
}





bool Display3D::isInViewport(int x, int y) {	
	return ((mpCamera->getLeft() < x) && (x < mpCamera->getRight()) &&
			(mpCamera->getBottom() < y) && (y < mpCamera->getTop()));
}

int Display3D::getHit(int bufSize, int xMouse, int yMouse, float delX, float delY, float distMax) {
	GLuint selectBuf[bufSize];
	GLint hits;
	
	glSelectBuffer(bufSize, selectBuf);
	(void) glRenderMode(GL_SELECT);
	glInitNames();
	//glPushName(0);	
	
	mpCamera->initPickingProjection(xMouse, yMouse, delX, delY);
	
	float farClip = mpCamera->getFarClip();
	mpCamera->setFarClip(distMax);
	mpCamera->perspective();
	mpCamera->setFarClip(farClip);
	
	mpCamera->initModelview();
	mpCamera->place();
	
	renderClickableObjects();
	glFlush();
	hits = glRenderMode(GL_RENDER);	

	//std::cout << "number of hits " << hits << std::endl;
	
	return processHits(hits, selectBuf);	
}

void Display3D::renderClickableObjects() {
	for(int i=0;i<entityTree.size();i++) {
		entityTree[i]->renderClickableObjects();
	}	
}

void Display3D::renderObjects() {	
	shaderType shader;
	
	switch(mRendering) {
		case FIXED_PIPELINE_RENDERING:
			shader = FIXED_PIPELINE;
			break;
			
		case PER_FRAG_LIGHT_RENDERING:
			shader = PER_FRAG_LIGHT;
			break;
			
	}
	
	for(int i=0;i<entityTree.size();i++) {
		entityTree[i]->render(lightKey, shader, mGroundColor, mSkyColor, mHemiAxis);
	}		
}

void Display3D::renderOccluders() {
	for(int i=0;i<entityTree.size();i++) {
		entityTree[i]->renderOccluders();
	}	
}


void Display3D::renderScene() {
	if(mpFog != NULL) {
		mpFog->activate();
	}
	activateLights();
	renderObjects();
	releaseLights();
	if(mpFog != NULL) {
		mpFog->release();
	}
	
	/*if(mpGraphics->getShadowMode() == SHADOW_MAP) {
		//ambiantPass();
		//TODO
	} else {
		activateLights();
		renderObjects();
		releaseLights();
		
	}*/	
	
}


void Display3D::ambiantPass() {
	for(int i=0;i<lightsP.size();i++) {
		lightsP[i]->activate(i);
	}
	renderObjects();
	for(int i=0;i<lightsP.size();i++) {
		lightsP[i]->release();
	}	
}

void Display3D::spotLightPass(int i) {
	mpShadowMap->setLight(lightsS[i]);
	mpShadowMap->beginUpdate(10.,mpCamera->getFarClip()*2.);
	renderOccluders();
	mpShadowMap->endUpdate();
	
	mpCamera->initProjection();
	mpCamera->initModelview();
	mpCamera->place();
		
	mpShadowMap->activate();
	lightsS[i]->activate(0);
	renderObjects();
	lightsS[i]->release();
	mpShadowMap->release();	
}

void Display3D::dirLightPass(int i) {
	//TODO
}


void Display3D::renderMirroredScene(int i) {
		
	glClear (GL_STENCIL_BUFFER_BIT);  	
	
	glEnable (GL_STENCIL_TEST);
	glStencilFunc (GL_ALWAYS, 1, 255);
	glStencilOp (GL_KEEP, GL_KEEP, GL_REPLACE);
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	glDepthMask (GL_FALSE);
	mirrorList[i]->render(lightKey, FIXED_PIPELINE, mGroundColor, mSkyColor, mHemiAxis);	
	
	glDepthMask (GL_TRUE);		
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	glStencilFunc (GL_EQUAL, 1, 255);
	glStencilOp (GL_KEEP, GL_KEEP, GL_KEEP);
	glCullFace (GL_FRONT);
	
	double matrixRefl[16];
	mirrorList[i]->getReflectionMatrix(matrixRefl);
	
	
	if((mpSky != NULL)&&(mpSkyPlane != NULL)) {
		glPushMatrix ();
    	glLoadIdentity();
		mpCamera->lookAtTheSky();
		glMultMatrixd (matrixRefl);
		
		glDisable (GL_DEPTH_TEST);
		glDepthMask (GL_FALSE);
		
		if((mRendering != FIXED_PIPELINE_RENDERING)&&(mpSky != NULL)) {
			shaderMana->activateSky();	
			mpSky->display();			
		}
		
		if(mpSkyPlane != NULL) {
			if(mRendering != FIXED_PIPELINE_RENDERING) {
				shaderMana->activateFixedPipeline();
			}
			
			if(mpFog != NULL) {
				mpFog->activate();
			}
			mpSkyPlane->display();
			if(mpFog != NULL) {
				mpFog->release();	
			}
		}
			
		glDepthMask (GL_TRUE);
		glEnable (GL_DEPTH_TEST);	
		glPopMatrix ();			
	}
	
	glPushMatrix ();
	glMultMatrixd (matrixRefl);
	
	glEnable (GL_CLIP_PLANE0);
	double eq[4];
	mirrorList[i]->getEquation (eq);
	glClipPlane (GL_CLIP_PLANE0, eq);
	
	renderScene();
	
	glDisable (GL_CLIP_PLANE0);
	glPopMatrix ();
	glDisable (GL_STENCIL_TEST);
	
	glCullFace (GL_BACK);
}

void Display3D::renderMirrorSurface(int i) {

	shaderType shader;

	switch(mRendering) {
		case FIXED_PIPELINE_RENDERING:
			shader = FIXED_PIPELINE;
			break;
			
		case PER_FRAG_LIGHT_RENDERING:
			shader = PER_FRAG_LIGHT;
			break;
	}
	
	activateLights();
	if(mpFog != NULL) {
		mpFog->activate();
	}
	mirrorList[i]->render(lightKey, shader, mGroundColor, mSkyColor, mHemiAxis);
	if(mpFog != NULL) {	
		mpFog->release();
	}
	releaseLights();
	
}


void Display3D::renderObjectsReflectionInTheSea() {

	float height = mpWaterPlane->getHeight();	
	
	glCullFace(GL_FRONT);
	
	mpFBO->bind();
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);	
	
	//glViewport(0, 0, 1024, 1024);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	mpCamera->perspective();

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();	
	mpCamera->place();
	
	glTranslatef(0.0, height, 0.0);
	glEnable(GL_CLIP_PLANE0);
	GLdouble equation[4] = {0.0, -1.0, 0.0, 0.0}; 
	glClipPlane(GL_CLIP_PLANE0, equation);
	glScalef( 1.0, -1.0, 1.0 ); 
	glTranslatef(0.0, -height, 0.0);	
	
	activateLights();
	renderObjects();
	releaseLights();
	
	glDisable(GL_CLIP_PLANE0);	
	
	//mpFBO->renderToTexture();
	mpFBO->unBind();
	glCullFace(GL_BACK);	
}


void Display3D::renderWater() {
	
	glEnable(GL_STENCIL_TEST);
	glStencilFunc(GL_ALWAYS, 0x0, 0xff);
	glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
	
	mpWaterPlane->drawSurface();
	
	//first, we render the sky reflection
	if(mpSky != NULL) {
		//renderReflectedSky();
	}	
	
	//then we render the objects reflection
	glStencilFunc(GL_NOTEQUAL, 0, ~0);
 
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	glDisable(GL_DEPTH_TEST);

	//shaderMana->activateFixedPipeline();
	glActiveTexture(GL_TEXTURE0);
	//glDisable (GL_LIGHTING);

	//glDisable (GL_CULL_FACE);
	
	mpFBO->getTexture()->activate();	
	glBegin(GL_QUADS);
        glTexCoord2f(0.0f, 0.0f);
        glVertex3f(-1.0, -1.0, -0.5);
 
        glTexCoord2f(1.0f, 0.0f);
        glVertex3f(1.0f, -1.0f, -0.5);
 
        glTexCoord2f(1.0f, 1.0f);
        glVertex3f(1.0, 1.0, -0.5);
	
        glTexCoord2f(0.0f, 1.0f);
        glVertex3f(-1.0, 1.0, -0.5);	
    glEnd();
	mpFBO->getTexture()->release();
	
	//glEnable (GL_LIGHTING);

	//shaderMana->activateSky();
glDisable(GL_BLEND);
	
	glDisable(GL_STENCIL_TEST); 
	glEnable(GL_DEPTH_TEST);
	
	glMatrixMode(GL_PROJECTION); 
	glPopMatrix();
	
}


void Display3D::renderSky() {
	glPushMatrix ();
    glLoadIdentity();
	mpCamera->lookAtTheSky();
	glDisable (GL_DEPTH_TEST);
	glDepthMask (GL_FALSE);
	
	shaderMana->activateSky();	
	mpSky->display();
	
	glDepthMask (GL_TRUE);
	glEnable (GL_DEPTH_TEST);	
	glPopMatrix ();	
}

void Display3D::renderSkyPlane() {
	
	glPushMatrix ();
    glLoadIdentity();	
	mpCamera->lookAtTheSky();
	glDisable (GL_DEPTH_TEST);	
	glDepthMask (GL_FALSE);
	
	if(mpFog != NULL) {
		mpFog->activate();
	}
	mpSkyPlane->display();
	if(mpFog != NULL) {
		mpFog->release();
	}
	glDepthMask (GL_TRUE);
	glEnable (GL_DEPTH_TEST);	
	glPopMatrix ();		
}

void Display3D::renderReflectedSky() {
	
	float height = mpWaterPlane->getHeight();	
		
	mpCamera->initModelview();
	mpCamera->lookAtTheSky();
	
	glTranslatef(0.0, height, 0.0);
	glScalef( 1.0, -1.0, 1.0 ); 
	glTranslatef(0.0, -height, 0.0);
	
	glDepthMask (GL_FALSE);
	shaderMana->activateSky();
			
	mpSky->display();
	
	glDepthMask (GL_TRUE);	
}

void Display3D::render(GraphicOptions *graphics) {
	
	if(mpWaterPlane != NULL) {
		renderObjectsReflectionInTheSea();
	}
	
	mpCamera->initProjection();
	mpCamera->initModelview();	
	mpCamera->place();

	if(graphics->isReflectionEnabled()) {
		for(int i=0;i<mirrorList.size();i++) {
			renderMirroredScene (i);
		}

		
		glClear (GL_STENCIL_BUFFER_BIT);  	

		glEnable (GL_STENCIL_TEST);
		glStencilFunc (GL_ALWAYS, 1, 255);
		glStencilOp (GL_KEEP, GL_KEEP, GL_INCR);	

		for(int i=0;i<mirrorList.size();i++) {	
			renderMirrorSurface (i);
		}
			
		glStencilFunc (GL_NOTEQUAL, 1, 255);
		glStencilOp (GL_KEEP, GL_KEEP, GL_KEEP);
	}
	
	if((mRendering != FIXED_PIPELINE_RENDERING)&&(mpSky != NULL)) {
		renderSky();
	}	

	if(mpSkyPlane != NULL) {
		if(mRendering != FIXED_PIPELINE_RENDERING) {
			shaderMana->activateFixedPipeline();
		}
		renderSkyPlane();
	}
	
	if(graphics->isReflectionEnabled()) {
		glDisable (GL_STENCIL_TEST);
	} else {
		for(int i=0;i<mirrorList.size();i++) {	
			renderMirrorSurface (i);
		}
	}
	
	renderScene();	

	if(mpWaterPlane != NULL) {
		renderWater();
	}

}

