/*
 * Newt-Puzzle
 * Copyright (C) Aude Jaulent 2008, 2009 <elderme@tuxfamily.org>
 * 
 * Newt-Puzzle is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Newt-Puzzle is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "entity.h"


Entity::Entity():
	mpMaterial(NULL), mpMesh(NULL), mCastShadow(false), mNumClick(-2), mVisible(true) {
}

Entity::Entity (Material *material, Mesh *mesh, bool castShadow): 
	mpMaterial(NULL), mpMesh(NULL), mCastShadow(castShadow), mNumClick(-2), mVisible(true) {
		
	setMaterial(material);
	setMesh(mesh);
}

Entity::~Entity() {
		
	for(std::map<int, Entity*>::iterator it = children.begin(); it!=children.end(); it++) {
		Entity *child = (*it).second;
		delete child; 
	}
	children.clear();
			
	if(mpMaterial != NULL) {
		mpMaterial->decNbEntities();
	}
	
	if(mpMesh != NULL) {
		mpMesh->decNbEntities();
	}	
}


Material* Entity::getMaterial(void) {
	return mpMaterial;
}

void Entity::setMaterial(Material* material) {
	if(mpMaterial != NULL) {
		mpMaterial->decNbEntities();
	}
	mpMaterial = material;
	
	if(mpMaterial != NULL) {
		mpMaterial->incNbEntities();
	}
}


Mesh* Entity::getMesh() {
	return mpMesh;
}

void Entity::setMesh(Mesh* mesh) {
	if(mpMesh != NULL) {
		mpMesh->decNbEntities();
	}
	mpMesh = mesh;
	if(mpMesh != NULL) {	
		mpMesh->incNbEntities();
	}
}


void Entity::addMaterialsInTheSet(std::set<Material*> &matSet) {
	if(mpMaterial != NULL) {
		matSet.insert(mpMaterial);
	}
	
	for(std::map<int, Entity*>::iterator it = children.begin(); it!=children.end(); it++) {
		((*it).second)->addMaterialsInTheSet(matSet);
	}
}

void Entity::addMeshesInTheSet(std::set<std::pair<Mesh*, int> > &meshSet, int nbTextures) {
	if(mpMaterial != NULL) {
		nbTextures = mpMaterial->getNumberOfTextures();
	}
	
	if(mpMesh != NULL) {
		meshSet.insert(std::make_pair(mpMesh, nbTextures));
	}
	
	for(std::map<int, Entity*>::iterator it = children.begin(); it!=children.end(); it++) {
		((*it).second)->addMeshesInTheSet(meshSet, nbTextures);
	}	
}


void Entity::setVisibility(bool visible) {
	mVisible = visible;
}

bool Entity::isVisible() {
	return mVisible;
}

int Entity::getNumClick() {
	return mNumClick;
}

void Entity::setNumClick(int numClick) {
	mNumClick = numClick;	
}

bool Entity::castShadow() {
	return mCastShadow;
}



void Entity::initialize(Material *fatherMaterial) {
	
	Material *material;
	
	if(fatherMaterial != NULL) {
		material = fatherMaterial;
	} else {
		if(mpMaterial != NULL) {
			material = mpMaterial;
		}
	}
	
	if((material != NULL)&&(mpMesh!=NULL))  {
		mpMesh->initialize(material->getNumberOfTextures());
	}

	for(std::map<int, Entity*>::iterator it = children.begin(); it!=children.end(); it++) {
		((*it).second)->initialize(material);
	}	
}


std::set<int> Entity::getMaterialKeys() {
	std::set<int> materialKeys;
	
	if(mpMaterial!=NULL) {
		materialKeys.insert(mpMaterial->getMaterialKey());
	}
	
	for(std::map<int, Entity*>::iterator it = children.begin(); it!=children.end(); it++) {
		std::set<int> sonMaterialKeys = ((*it).second)->getMaterialKeys();
		materialKeys.insert(sonMaterialKeys.begin(), sonMaterialKeys.end());
	}	
	return materialKeys;
}


void Entity::render(int lightKey, shaderType shader, Color groundColor, Color skyColor, Vector3 hemiAxis) {
	
	if(mpMaterial!=NULL) {
		mpMaterial->activate(lightKey, shader);
		if(shader != FIXED_PIPELINE) {
			shaderMana->setAmbientColors (groundColor, skyColor, hemiAxis);	
		}
		drawGeometry(mpMaterial->getNumberOfTextures());
		mpMaterial->release();
	} else {
		glPushMatrix();
		applyTransformation();
		for(std::map<int, Entity*>::iterator it = children.begin(); it!=children.end(); it++) {
			((*it).second)->render(lightKey, shader, groundColor, skyColor, hemiAxis);
		}
		glPopMatrix();		
	}
}

void Entity::drawGeometry(int nbTextures) {
	glPushMatrix();
	applyTransformation();
	if((mVisible)&&(mpMesh!=NULL)) {	
		mpMesh->draw(nbTextures);
	}
	for(std::map<int, Entity*>::iterator it = children.begin(); it!=children.end(); it++) {
		((*it).second)->drawGeometry(nbTextures);
	}
	glPopMatrix();	
}

void Entity::renderOccluders() {
	if(mVisible) {	
		glPushMatrix();
		applyTransformation();
		if(mpMesh!=NULL) {
			mpMesh->draw(0);
		}
		for(std::map<int, Entity*>::iterator it = children.begin(); it!=children.end(); it++) {
			((*it).second)->renderOccluders();
		}
		glPopMatrix();
	}
}

void Entity::renderClickableObjects() {
	
	if(mNumClick > -2) {
		if(mNumClick >= 0) {			
			glPushName(mNumClick);
			glPushMatrix();
			applyTransformation();
			if(mpMesh!=NULL) {
				mpMesh->draw(0);
			}
			
			for(std::map<int, Entity*>::iterator it = children.begin(); it!=children.end(); it++) {
				((*it).second)->renderClickableObjects();
			}
			glPopMatrix();
			glPopName();
			
		} else {
			
			glPushMatrix();
			applyTransformation();	
			for(std::map<int, Entity*>::iterator it = children.begin(); it!=children.end(); it++) {
				((*it).second)->renderClickableObjects();
			}
			glPopMatrix();	
			
		}
	}
}


bool Entity::isAChild(int key) {
	return (children.find(key) != children.end());
}


void Entity::addChild(Entity *child) {
	// we find a free negative key
	int key = -1;
	while(children.find(key) != children.end()) {
		key--;
	}
	children[key] = child;
}

void Entity::addChild(Entity *child, int key) {
	children[key] = child;	
}

void Entity::removeChild(int key) {
	delete children[key];
	children.erase(key);	
}

void Entity::removeAllChildren() {
	for(std::map<int, Entity*>::iterator it = children.begin(); it!=children.end(); it++) {
		Entity* child = (*it).second;
		delete child;
	}	
	children.erase(children.begin(),children.end());
}


/****************************************/
/*		EntityWithPosition				*/
/****************************************/

EntityWithPosition::EntityWithPosition() : mCenter(Vector3 (0.,0.,0.)), mRotAxis(Vector3(1.,0.,0.)), mAngle(0.) {
}

EntityWithPosition::EntityWithPosition(Vector3 center, Vector3 rotAxis, float angle):
	Entity(), mCenter(center), mRotAxis(rotAxis), mAngle(angle) {

}


EntityWithPosition::EntityWithPosition(Material *material, Mesh *mesh, bool castShadow, Vector3 center, Vector3 rotAxis, float angle):
	Entity(material, mesh, castShadow), mCenter(center), mRotAxis(rotAxis), mAngle(angle) {

}

EntityWithPosition::~EntityWithPosition() {
	
}

void EntityWithPosition::setPosition(Vector3 center) {
	mCenter = center;	
}

void EntityWithPosition::setPosition(Vector3 center, Vector3 rotAxis, float angle) {
	mCenter = center;
	mRotAxis = rotAxis;
	mAngle = angle;
}


void EntityWithPosition::applyTransformation() {
	glTranslatef(mCenter.x, mCenter.y, mCenter.z);
	glRotatef(mAngle, mRotAxis.x, mRotAxis.y, mRotAxis.z);
}



